/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include <qdatetime.h>
#include <qmainwindow.h>
#include <qstatusbar.h>
#include <qmessagebox.h>
#include <qmenubar.h>
#include <qapplication.h>
#include <qkeycode.h>
#include <qpainter.h>
#include <qlabel.h>
#include <qimage.h>
#include <pixmaploader.h>

#include "canvas.h"

#include <stdlib.h>

static const int imageRTTI = 984376;


class ImageItem: public QCanvasRectangle
{
public:
    ImageItem( QImage img, QCanvas *canvas );
    int rtti () const { return imageRTTI; }
    bool hit( const QPoint&) const;
protected:
    void drawShape( QPainter & );
private:
    QImage image;
};


ImageItem::ImageItem( QImage img, QCanvas *canvas )
    : QCanvasRectangle( canvas ), image(img)
{
    setSize( image.width(), image.height() );
}


void ImageItem::drawShape( QPainter &p )
{
    p.drawImage( x(), y(), image );
}

bool ImageItem::hit( const QPoint &p ) const 
{ 
    int ix = p.x()-int(x());
    int iy = p.y()-int(y());
    if ( !image.valid( ix , iy ) )
	return FALSE;
    QRgb pixel = image.pixel( ix, iy );
    return qAlpha( pixel ) != 0;
}

FigureEditor::FigureEditor(
	QCanvas& c, QWidget* parent,
	const char* name, WFlags f) :
    QCanvasView(&c,parent,name,f)
{
}

void FigureEditor::contentsMousePressEvent(QMouseEvent* e)
{
    QCanvasItemList l=canvas()->collisions(e->pos());
    for (QCanvasItemList::Iterator it=l.begin(); it!=l.end(); ++it) {
	if ( (*it)->rtti() == imageRTTI ) {
	    ImageItem *item= (ImageItem*)(*it);
	    if ( !item->hit( e->pos() ) )
		 continue;
	}
	moving = *it;
	moving_start = e->pos();
	return;
    }
    moving = 0;
}

void FigureEditor::clear()
{
    QCanvasItemList list = canvas()->allItems();
    QCanvasItemList::Iterator it = list.begin();
    for (; it != list.end(); ++it) {
	if ( *it )
	    delete *it;
    }
}

void CanvasWindow::init()
{
    clear();

    static int r=24;
    srand48(++r);

    int i;

    for ( i=0; i<canvas.width() / 56; i++) {
	addButterfly();
    }
    for ( i=0; i<canvas.width() / 85; i++) {
	addHexagon();
    }
}

void FigureEditor::contentsMouseMoveEvent(QMouseEvent* e)
{
    if ( moving ) {
	moving->moveBy(e->pos().x() - moving_start.x(),
		       e->pos().y() - moving_start.y());
	moving_start = e->pos();
	canvas()->update();
    }
}




CanvasWindow::CanvasWindow(QWidget* parent, const char* name, WFlags f) :
    QMainWindow(parent,name,f),
    canvas( 230, 270 )
{
    setCaption( "Canvas Demo" );

    canvas.setAdvancePeriod( 50 );

    editor = new FigureEditor(canvas,this);
    QMenuBar* menu = menuBar();

    QPopupMenu* file = new QPopupMenu;
    file->insertItem("Fill canvas", this, SLOT(init()));
    file->insertItem("Erase canvas", this, SLOT(clear()));
    file->insertSeparator();
    file->insertItem("Quit", this, SLOT(close()));
    menu->insertItem("File", file);

    QPopupMenu* edit = new QPopupMenu;
    edit->insertItem("Add Circle", this, SLOT(addCircle()));
    edit->insertItem("Add Hexagon", this, SLOT(addHexagon()));
    edit->insertItem("Add Polygon", this, SLOT(addPolygon()));
    edit->insertItem("Add Line", this, SLOT(addLine()));
    edit->insertItem("Add Rectangle", this, SLOT(addRectangle()));
    edit->insertItem("Add Alpha-blended image", this, SLOT(addButterfly()));
    edit->insertSeparator();
    edit->insertItem("Enlarge", this, SLOT(enlarge()));
    edit->insertItem("Shrink", this, SLOT(shrink()));
    menu->insertItem("Edit", edit);

    options = new QPopupMenu;
    dbf_id = options->insertItem("Double buffer", this, SLOT(toggleDoubleBuffer()));
    options->setItemChecked(dbf_id, TRUE);
    menu->insertItem("Options",options);

    QPopupMenu* help = new QPopupMenu;
    help->insertItem("Help...", this, SLOT(help()));
    help->setItemChecked(dbf_id, TRUE);
    menu->insertItem("Help",help);

    setCentralWidget(editor);

    init();
}

CanvasWindow::~CanvasWindow()
{
}

void CanvasWindow::clear()
{
    editor->clear();
}

void CanvasWindow::help()
{
    static QMessageBox* about = new QMessageBox( "Qt Canvas Example",
	    "The QCanvas classes example\nwith alpha blending",
	    QMessageBox::Information, 1, 0, 0, this, 0, FALSE );
    about->setButtonText( 1, "Dismiss" );
    about->show();
}

void CanvasWindow::aboutQt()
{
    QMessageBox::aboutQt( this, "Qt Canvas Example" );
}

void CanvasWindow::toggleDoubleBuffer()
{
    bool s = !options->isItemChecked(dbf_id);
    options->setItemChecked(dbf_id,s);
    canvas.setDoubleBuffering(s);
}

void CanvasWindow::enlarge()
{
    canvas.resize(canvas.width()*4/3, canvas.height()*4/3);
}

void CanvasWindow::shrink()
{
    canvas.resize(canvas.width()*3/4, canvas.height()*3/4);
}

void CanvasWindow::addButterfly()
{
    static QImage *img = 0;
    if ( !img ) {
	img = new QImage;
	img->load( PixmapLoader::findPixmap( "butterfly" ) );
    }
    QCanvasPolygonalItem* i = new ImageItem(*img,&canvas);
    i->move(lrand48()%(canvas.width()-img->width()),
	    lrand48()%(canvas.height()-img->height()));
    i->setZ(lrand48()%256+250);
    i->show();
}


void CanvasWindow::addCircle()
{
    QCanvasPolygonalItem* i = new QCanvasEllipse(50,50,&canvas);
    i->setBrush( QColor(lrand48()%32*8,lrand48()%32*8,lrand48()%32*8) );
    i->move(lrand48()%canvas.width(),lrand48()%canvas.height());
    i->setZ(lrand48()%256);
    i->show();
}

void CanvasWindow::addHexagon()
{
    QCanvasPolygon* i = new QCanvasPolygon(&canvas);
    const int size = canvas.width() / 25;
    QPointArray pa(6);
    pa[0] = QPoint(2*size,0);
    pa[1] = QPoint(size,-size*173/100);
    pa[2] = QPoint(-size,-size*173/100);
    pa[3] = QPoint(-2*size,0);
    pa[4] = QPoint(-size,size*173/100);
    pa[5] = QPoint(size,size*173/100);
    i->setPoints(pa);
    i->setBrush( QColor(lrand48()%32*8,lrand48()%32*8,lrand48()%32*8) );
    i->move(lrand48()%canvas.width(),lrand48()%canvas.height());
    i->setZ(lrand48()%256);
    i->show();
}

void CanvasWindow::addPolygon()
{
    QCanvasPolygon* i = new QCanvasPolygon(&canvas);
    const int size = canvas.width()/3;
    QPointArray pa(6);
    pa[0] = QPoint(0,0);
    pa[1] = QPoint(size,size/5);
    pa[2] = QPoint(size*4/5,size);
    pa[3] = QPoint(size/6,size*5/4);
    pa[4] = QPoint(size*3/4,size*3/4);
    pa[5] = QPoint(size*3/4,size/4);
    i->setPoints(pa);
    i->setBrush( QColor(lrand48()%32*8,lrand48()%32*8,lrand48()%32*8) );
    i->move(lrand48()%canvas.width(),lrand48()%canvas.height());
    i->setZ(lrand48()%256);
    i->show();
}

void CanvasWindow::addLine()
{
    QCanvasLine* i = new QCanvasLine(&canvas);
    i->setPoints( lrand48()%canvas.width(), lrand48()%canvas.height(),
                  lrand48()%canvas.width(), lrand48()%canvas.height() );
    i->setPen( QColor(lrand48()%32*8,lrand48()%32*8,lrand48()%32*8) );
    i->setZ(lrand48()%256);
    i->show();
}

void CanvasWindow::addRectangle()
{
    QCanvasPolygonalItem *i = new QCanvasRectangle( lrand48()%canvas.width(),lrand48()%canvas.height(),
			    canvas.width()/5,canvas.width()/5,&canvas);
    int z = lrand48()%256;
    i->setBrush( QColor(z,z,z) );
    i->setZ(z);
    i->show();
}

