/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <qmessagebox.h>
#include <qdir.h>
#include <qheader.h>
#include <qtoolbar.h>
#include <qpopupmenu.h>
#include <qmenubar.h>
#include <qaction.h>
#include <qstringlist.h>
#include <qcursor.h>
#include <qmultilineedit.h>
#include <qfont.h>
#include <pixmaploader.h>

#include "inlineedit.h"
#include "filetypeinfodb.h"
#include "filebrowser.h"

//
//  FileItem
//
FileItem::FileItem( QListView * parent, const QFileInfo & fi )
    : QListViewItem( parent ),
      fileInfo( fi )
{
    FileTypeInfoDB *  ft = FileTypeInfoDB::instance();
    QDate d = fi.lastModified().date();
    
    setText( 0, fi.fileName() );
    setText( 1, QString().sprintf("%02d-%02d-%4d",d.day(), d.month(), 
				  d.year() ) );
    setText( 2, QString().setNum( fi.size() ) );
        
    if( fi.isDir() )
	setText( 3, "directory" );
    else if( isLib() )
	setText( 3, "library" );
    else
	setText( 3, ft->type( fi.filePath()) );

    if( fi.isDir() ){
	if( !QDir( fi.filePath() ).isReadable() )
	    setPixmap( 0, PixmapLoader::loadPixmap( "lockedfolder" ) );
	else
	    setPixmap( 0, PixmapLoader::loadPixmap( "folder" ) );
    } else if( !fi.isReadable() ){
	setPixmap( 0, PixmapLoader::loadPixmap( "locked" ) );
    } else if( fi.isExecutable() && !isLib() ){
	if( ft->type( fi.filePath() ) != "application/octet-stream" )
	    setPixmap( 0 , 
		       PixmapLoader::loadPixmap( ft->pixmap( fi.fileName() )));
	else
	    setPixmap( 0 , PixmapLoader::loadPixmap( "exec" ) );
    } else if( isLib() ){
	setPixmap( 0 , PixmapLoader::loadPixmap( "library" ) );
    } else {
	setPixmap( 0 , PixmapLoader::loadPixmap( ft->pixmap( fi.fileName() )));
    }
}

QString FileItem::key( int column, bool ascending ) const
{
    QString tmp;
    
    ascending = ascending;
    
    if( (column == 0) && fileInfo.isDir() ){ // Sort by name
	// We want the directories to appear at the top of the list
	tmp = (char) 0;
	return (tmp + text( column ).lower()); 
    } else if( column == 1 ) { // Sort by date
	QDateTime epoch( QDate( 1980, 1, 1 ) );
        tmp.sprintf( "%08d", epoch.secsTo( fileInfo.lastModified() ) );
	return tmp;
    } else if( column == 2 ) { // Sort by size
	return tmp.sprintf( "%08d", fileInfo.size() );
    }
    
    return text( column ).lower();
}

bool FileItem::isLib()
{
    // This is of course not foolproof
    if( !qstrncmp("lib", fileInfo.baseName(), 3) && 
	(fileInfo.extension().contains( "so" ) ||
	 fileInfo.extension().contains( "a" )) ) 
	return TRUE;
    else
	return FALSE;
}

int FileItem::launch()
{
    FileTypeInfoDB * ft = FileTypeInfoDB::instance();
    QString app = ft->appname( text( 0 ) ), cmdStr;

    if( app.length() > 0 ){
	cmdStr.sprintf( "%s \"%s\"", (const char *) app, 
			(const char *) fileInfo.filePath() );
    } else if( fileInfo.isExecutable() && !isLib() ){
	cmdStr = fileInfo.filePath();
    } else {
	return 0;
    }
#ifndef SINGLE_APP
    const char * argv[4];
    int pid = fork();
	
    if( pid == -1 ) return -1;
    if( pid == 0 ) {	    
	argv[0] = "sh";
	argv[1] = "-c";
	argv[2] = cmdStr;
	argv[3] = 0;
	execv( "/bin/sh", (char * const *) argv );
	exit( 0 );
    }
#endif
    return 1;
}

bool FileItem::rename( const QString & name )
{
    QString oldpath, newpath;
    
    oldpath = fileInfo.filePath();
    newpath = fileInfo.dirPath() + "/" + name;

    if( ::rename( (const char *) oldpath, (const char *) newpath ) != 0 )
	return FALSE;
    else
	return TRUE;
}

//
//  FileView
//
FileView::FileView( const QString & dir, QWidget * parent, 
		    const char * name )
    : QListView( parent, name ),
      menuTimer( this ), 
      le( NULL ), 
      itemToRename( NULL )
{
    addColumn( "Name" );
    addColumn( "Date" );
    addColumn( "Size" );
    addColumn( "Type" );
    
    setColumnAlignment( 2, AlignRight );
   /*    setColumnWidthMode( 0, Manual );
	 setColumnWidthMode( 1, Manual );
	 setColumnWidthMode( 2, Manual );
	 setColumnWidthMode( 3, Manual );
	 setColumnWidth( 0, 100 );
	 setColumnWidth( 1, 65 );
	 setColumnWidth( 2, 50 );
	 setColumnWidth( 3, 0 ); */
    
    setMultiSelection( TRUE );

    header()->hide();
    
    generateDir( dir );
    textViewer = 0;
    
    connect( this, SIGNAL( clicked( QListViewItem * )), 
	     SLOT( itemClicked( QListViewItem * )) );
    connect( this, SIGNAL( doubleClicked( QListViewItem * )), 
	     SLOT( itemDblClicked( QListViewItem * )) );
    connect( this, SIGNAL( selectionChanged() ), SLOT( cancelMenuTimer() ) );
    connect( &menuTimer, SIGNAL( timeout() ), SLOT( showFileMenu() ) );
}


void FileView::updateDir()
{
    generateDir( currentDir );
}

void FileView::setDir( const QString & dir )
{
    dirHistory += currentDir;  
    generateDir( dir );
}

void FileView::generateDir( const QString & dir )
{    
    QDir d( dir );

    if( d.exists() && !d.isReadable() ) return;
        
    currentDir = d.canonicalPath();
	
    d.setFilter( QDir::Dirs | QDir::Files | QDir::NoSymLinks );
    d.setSorting( QDir::Name | QDir::DirsFirst | QDir::IgnoreCase |
	          QDir::Reversed );
	
    const QFileInfoList * list = d.entryInfoList();
    QFileInfoListIterator it( *list );
    QFileInfo *fi;

    clear();
    while( (fi = it.current()) ){
	if( (fi->fileName() == ".") || (fi->fileName() == "..") ){
	    ++it;
	    continue;
	}
	(void) new FileItem( (QListView *) this, *fi );
	++it;
    }
    
    emit dirChanged();
}

void FileView::rename()
{
    itemToRename = (FileItem *) currentItem();
    const QPixmap * pm;
    int pmw;
    
    if( itemToRename == NULL ) return;
    
    if( ( pm = itemToRename->pixmap( 0 ) ) == NULL )
	pmw = 0;
    else 
	pmw = pm->width();

    ensureItemVisible( itemToRename );
    horizontalScrollBar()->setValue( 0 );
    horizontalScrollBar()->setEnabled( FALSE );
    verticalScrollBar()->setEnabled( FALSE );
    
    selected = isSelected( itemToRename );
    setSelected( itemToRename, FALSE );
    
    if( le == NULL ){
	le = new InlineEdit( this );
	le->setFrame( FALSE );
	connect( le, SIGNAL( lostFocus() ), SLOT( endRenaming() ) ); 
    }

    QRect r = itemRect( itemToRename );
    r.setTop( r.top() + frameWidth() + 1 );
    r.setLeft( r.left() + frameWidth() + pmw );
    r.setBottom( r.bottom() + frameWidth() );
    r.setWidth( columnWidth( 0 ) - pmw );
        
    le->setGeometry( r );
    le->setText( itemToRename->text( 0 ) );
    le->selectAll();
    le->show();
    le->setFocus();
}

void FileView::endRenaming()
{
    if( le && itemToRename ){
	le->hide();
	setSelected( itemToRename, selected );
	
	if( !itemToRename->rename( le->text() ) ){
	    QMessageBox::warning( this, tr( "Rename file" ), 
				  tr( "Rename failed!" ), tr( "&Ok" ) );
	} else {
	    updateDir();
	}
	itemToRename = NULL;
	horizontalScrollBar()->setEnabled( TRUE );
	verticalScrollBar()->setEnabled( TRUE );
    }
}

void FileView::copy()
{
    FileItem * i;
    
    if((i = (FileItem *) firstChild()) == 0) return;    
    
    flist.clear();
    while( i ){
	if( i->isSelected() /*&& !i->isDir()*/ ){
	    flist += i->getFilePath();
	}
	i = (FileItem *) i->nextSibling();
    }
}

void FileView::paste()
{
    int i, err;
    QString cmd, dest, basename, cd = currentDir;    

    if(cd == "/") cd = "";
   
    for ( QStringList::Iterator it = flist.begin(); it != flist.end(); ++it ) {
	basename = (*it).mid((*it).findRev("/") + 1, (*it).length());
	
	dest = cd + "/" + basename;
	if( QFile( dest ).exists() ){
	    i = 1;
	    dest = cd + "/Copy of " + basename;
	    while( QFile( dest ).exists() ){
		dest.sprintf( "%s/Copy (%d) of %s", (const char *) cd, i++, 
			      (const char *) basename );
	    }
	}
	
	//
	// Copy a directory recursively using the "cp" command -
	// may have to be changed
	//
	if( QFileInfo( (*it) ).isDir() ){
	    cmd = "/bin/cp -fpR \"" + (*it) +"\" " + "\"" + dest + "\"";
	    err = system( (const char *) cmd );
	} else if( !copyFile( dest, (*it) ) ){
	    err = -1;
	} else {
	    err = 0;
	}

	if( ( err < 0 ) || ( err == 127 ) ){
	    QMessageBox::warning( this, tr("Paste file"), tr("Paste failed!"),
				  tr("&Ok") );
	    break;
	} else {
	    updateDir();
	    QListViewItem * i = firstChild();
	    basename = dest.mid( dest.findRev("/") + 1, dest.length() );
	
	    while( i ){
		if( i->text(0) == basename ){
		    setCurrentItem( i );
		    ensureItemVisible( i ); 
		    break;
		}
		i = i->nextSibling();
	    }
	}
    }
}

bool FileView::copyFile( const QString & dest, const QString & src )
{
    char bf[ 50000 ];
    int  bytesRead;
    bool success = TRUE;
    struct stat status;
    
    QFile s( src );
    QFile d( dest );
    
    if( s.open( IO_ReadOnly | IO_Raw ) && 
	d.open( IO_WriteOnly | IO_Raw ) )
    {
	while( (bytesRead = s.readBlock( bf, sizeof( bf ) )) == 
	       sizeof( bf ) )
	{
	    if( d.writeBlock( bf, sizeof( bf ) ) != sizeof( bf ) ){
		success = FALSE;
		break;
	    }
	}
	if( success && (bytesRead > 0) ){
	    d.writeBlock( bf, bytesRead );
	}
    } else {
	success = FALSE;
    }
    
    // Set file permissions    
    if( stat( (const char *) src, &status ) == 0 ){
	chmod( (const char *) dest, status.st_mode );
    }
    
    return success;
}

void FileView::del()
{
    FileItem * i;
    QStringList fl;
    QString cmd;
    int err;
        
    if((i = (FileItem *) firstChild()) == 0) return;    
    
    while( i ){
	if( i->isSelected() ){
	    fl += i->getFilePath();
	}
	i = (FileItem *) i->nextSibling();
    }
    if( fl.count() < 1 ) return;
    
    if( QMessageBox::warning( this, tr("Cut"), tr("Are you sure?"), 
			      tr("&Yes"), tr("&No") ) == 0)
    {
	//
	// Dependant upon the "rm" command - will probably have to be replaced
	//
	for ( QStringList::Iterator it = fl.begin(); it != fl.end(); ++it ) {
	    cmd = "/bin/rm -rf \"" + (*it) + "\"";	    
	    err = system( (const char *) cmd );
	    if( ( err < 0 ) || ( err == 127 ) ){
		QMessageBox::warning( this, tr("Cut"), tr("Cut failed!"),
				      tr("&Ok") );
		break;
	    }
	}
	updateDir();
    }
}

void FileView::newFolder()
{
    int t = 1;
    FileItem * i;
    QString nd = currentDir + "/NewFolder";
	
    while( QFile( nd ).exists() ){
	nd.sprintf( "%s/NewFolder (%d)", (const char *) currentDir, t++ );
    }
    
    if( mkdir( (const char *) nd, 0777 ) != 0){
	QMessageBox::warning( this, tr( "New folder" ), 
			      tr( "Folder creation failed!" ), 
			      tr( "&Ok" ) );
	return;
    }
    updateDir();
    
    if((i = (FileItem *) firstChild()) == 0) return;    
    
    while( i ){
	if( i->isDir() && ( i->getFilePath() == nd ) ){
	    setCurrentItem( i );
	    rename();
	    break;
	}
	i = (FileItem *) i->nextSibling();
    }
}

void FileView::closeTextView()
{
    if( textViewer ){
    /*
	delete textViewer;
	textViewer = 0;
    */
	textViewer->setText( "" );
	textViewer->hide();
	emit textViewDeactivated();
    }
}

void FileView::viewAsText()
{
    FileItem * i = (FileItem *) currentItem();
    QString text;
    char bf[256];
    int res;
    
    if( !i ) return;
    
    QFile f( i->getFilePath() );
    if( f.open( IO_ReadOnly | IO_Raw ) ){
	while( (res = f.readBlock( bf, 255 )) > 0 ){
	    if( res > 0 ){
		bf[ res ] = 0;
		text += bf;
	    } else {
		break;
	    }
	}	
	if ( !textViewer )
	    textViewer = new QMultiLineEdit( parentWidget() );
	textViewer->setText( text );
	textViewer->setReadOnly( TRUE );
	emit textViewActivated( textViewer );
    } else {
	QMessageBox::warning( this, tr( "View as text" ),
			      tr( "Unable to open file!" ), tr( "&Ok" ) );
    }
}

void FileView::itemClicked( QListViewItem * i)
{
    FileItem * t = (FileItem *) i;
    
    if( t == NULL ) return;
    if( t->isDir() ){
	setDir( t->getFilePath() );
    }
}

void FileView::itemDblClicked( QListViewItem * i)
{
    FileItem * t = (FileItem *) i;
    
    if(t == NULL) return;
    if(t->launch() == -1){
	QMessageBox::warning( this, tr( "Launch Application" ),
			      tr( "Launch failed!" ), tr( "&Ok" ) );
    }
}

void FileView::parentDir()
{
    setDir( currentDir + "./.." );
}

void FileView::lastDir()
{
    if( dirHistory.count() == 0 ) return;
    
    generateDir( dirHistory.last() );
    dirHistory.remove( dirHistory.last() );
}
  
void FileView::contentsMousePressEvent( QMouseEvent * e )
{
    QListView::contentsMousePressEvent( e );
    menuTimer.start( 750, TRUE );
}

void FileView::contentsMouseReleaseEvent( QMouseEvent * e )
{
    QListView::contentsMouseReleaseEvent( e );
    menuTimer.stop();
}

void FileView::cancelMenuTimer()
{
    if( menuTimer.isActive() )
	menuTimer.stop();
}

void FileView::run()
{
    FileItem * i = (FileItem *) currentItem();
    i->launch();
}

void FileView::showFileMenu()
{
    FileItem * i = (FileItem *) currentItem();
    QPopupMenu * m = new QPopupMenu( this );
    FileTypeInfoDB * ft = FileTypeInfoDB::instance();
    QString associatedApp = ft->appname( i->text( 0 ) );
    
    if( associatedApp.length() > 0 ){
	m->insertItem( PixmapLoader::loadPixmap( ft->pixmap( associatedApp ) ),
		       tr( "&Open in " + associatedApp ), this, 
		       SLOT( run() ) );
	m->insertSeparator();	
    } else if( i->isExecutable() && !i->isDir() ){
	m->insertItem( PixmapLoader::loadPixmap( i->text( 0 ) ),
		       tr( "&Run" ), this, SLOT( run() ) );
	m->insertSeparator();
    }

    if( !i->isLib() /*&& !i->isExecutable()*/ && !i->isDir() ){
	m->insertItem( PixmapLoader::loadPixmap( "txt.xpm" ),
		       tr( "&View as text" ), this, SLOT( viewAsText() ) );
    }
    
    m->insertItem( tr( "R&ename" ), this, SLOT( rename() ) );
    m->insertItem( PixmapLoader::loadPixmap("copy"), 
		   tr( "&Copy" ), this, SLOT( copy() ) );
    m->insertItem( PixmapLoader::loadPixmap("paste"),
		   tr( "&Paste" ), this, SLOT( paste() ) );
    m->insertItem( PixmapLoader::loadPixmap("cut"),
		   tr( "&Cut" ), this, SLOT( del() ) );
    m->insertSeparator();
    m->insertItem( tr( "&Select all" ), this, SLOT( selectAll() ) );
    m->insertItem( tr( "Dese&lect all" ), this, SLOT( deselectAll() ) );
    m->popup( QCursor::pos() );
}

//
//  FileBrowser
//

FileBrowser::FileBrowser( QWidget * parent, 
			  const char * name, WFlags f ) :
    QMainWindow( parent, name, f )
{
    init(QDir::current().canonicalPath());
}

FileBrowser::FileBrowser( const QString & dir, QWidget * parent, 
			  const char * name, WFlags f ) :
    QMainWindow( parent, name, f )
{
    init(dir);
}

void FileBrowser::init(const QString & dir)
{
    setCaption( tr("File Browser") );
    fileView = new FileView( dir, this );
    fileView->setAllColumnsShowFocus( TRUE );
    
    setCentralWidget( fileView );
    setToolBarsMovable( FALSE );
    
    toolBar = new QToolBar( this );
    toolBar->setHorizontalStretchable( TRUE );
    
    menuBar = new QMenuBar( toolBar );
    
    dirMenu = new QPopupMenu( this );
    menuBar->insertItem( tr( "&Dir" ), dirMenu );

    sortMenu = new QPopupMenu( this );
    menuBar->insertItem( tr( "&Sort" ), sortMenu );
    sortMenu->insertItem( tr( "by &Name "), this, SLOT( sortName() ) );
    sortMenu->insertItem( tr( "by &Date "), this, SLOT( sortDate() ) );
    sortMenu->insertItem( tr( "by &Size "), this, SLOT( sortSize() ) );
    sortMenu->insertItem( tr( "by &Type "), this, SLOT( sortType() ) );
    sortMenu->setItemChecked( sortMenu->idAt(0), TRUE );
    
    tbRegular  = new QToolBar( this );
    tbTextView = new QToolBar( this );
    
    QAction * a = new QAction( tr("Close"), 
		     PixmapLoader::loadPixmap( "close" ),
		     QString::null, 0, this, 0 );
    connect( a, SIGNAL( activated() ), fileView, SLOT( closeTextView() ) );
    a->addTo( tbTextView );
    tbTextView->hide();
    
    a = new QAction( tr("Previous dir"), 
		     PixmapLoader::loadPixmap( "back.xpm" ),
		     QString::null, 0, this, 0 );
    connect( a, SIGNAL( activated() ), fileView, SLOT( lastDir() ) );
    a->addTo( tbRegular );

    a = new QAction( tr("Parent dir"), 
			       PixmapLoader::loadPixmap( "up.xpm" ),
			       QString::null, 0, this, 0 );
    connect( a, SIGNAL( activated() ), fileView, SLOT( parentDir() ) );
    a->addTo( tbRegular );

    a = new QAction( tr("New folder"), 
		     PixmapLoader::loadPixmap( "new" ),
		     QString::null, 0, this, 0 );
    connect( a, SIGNAL( activated() ), fileView, SLOT( newFolder() ) );
    a->addTo( tbRegular );

    a = new QAction( tr("Copy"), 
		     PixmapLoader::loadPixmap( "copy" ),
		     QString::null, 0, this, 0 );
    connect( a, SIGNAL( activated() ), fileView, SLOT( copy() ) );
    a->addTo( tbRegular );

    pasteAction = new QAction( tr("Paste"), 
		     PixmapLoader::loadPixmap( "paste" ),
		     QString::null, 0, this, 0 );
    connect( pasteAction, SIGNAL( activated() ), fileView, SLOT( paste() ) );
    pasteAction->addTo( tbRegular );
    // pasteAction->setEnabled( false );

    a = new QAction( tr("Cut"), 
		     PixmapLoader::loadPixmap( "cut" ),
		     QString::null, 0, this, 0 );
    connect( a, SIGNAL( activated() ), fileView, SLOT( del() ) );
    a->addTo( tbRegular );
    
    connect( fileView, SIGNAL( dirChanged() ), SLOT( updateDirMenu() ) );

    connect( fileView, SIGNAL( textViewActivated( QWidget * ) ), 
	     SLOT( setTextViewMode( QWidget * ) ) );
    connect( fileView, SIGNAL( textViewDeactivated() ), 
	     SLOT( setRegularMode() ) );
    updateDirMenu();
}

void FileBrowser::setTextViewMode( QWidget * w )
{
    if( w != NULL )
	setCentralWidget( w );

    tbRegular->hide();
    tbTextView->show();
    // ## For some reason the toolbars are resized incorrectly if I
    // ## use hide() instead of enable/disable them. 
    //menuBar->hide();
    menuBar->setEnabled( FALSE );
}

void FileBrowser::setRegularMode()
{
    setCentralWidget( fileView );
    tbTextView->hide();
    tbRegular->show();
    // ## For some reason the toolbars are resized incorrectly if I
    // ## use hide()/show() instead of enable/disable them. 
    //menuBar->show();
    menuBar->setEnabled( TRUE );
}

void FileBrowser::dirSelected( int id )
{
    int i = 0, j;
    QString dir;
    
    // Bulid target dir from menu
    while( (j = dirMenu->idAt( i )) != id ){
	dir += dirMenu->text( j ).stripWhiteSpace();
	if( dirMenu->text( j ) != "/" ) dir += "/";     
	i++;
    }
    dir += dirMenu->text( dirMenu->idAt( i ) ).stripWhiteSpace();
    
    fileView->setDir( dir );
}

void FileBrowser::updateDirMenu()
{
    QString spc, cd = fileView->cd();
    QStringList l = QStringList::split( "/", cd );
    int i = 0;
    
    dirMenu->clear();
    dirMenu->insertItem( tr( "/" ), this, SLOT( dirSelected(int) ) );
		     
    for ( QStringList::Iterator it = l.begin(); it != l.end(); ++it ) {
	spc.fill( ' ', i++);
	dirMenu->insertItem( spc + (*it), this, 
			     SLOT( dirSelected(int) ) );
    }
    dirMenu->setItemChecked( dirMenu->idAt( l.count() ), TRUE );
}

void FileBrowser::sortName()
{
    fileView->setSorting( 0 );
    fileView->sort();
    sortMenu->setItemChecked( sortMenu->idAt( 0 ), TRUE );
    sortMenu->setItemChecked( sortMenu->idAt( 1 ), FALSE );
    sortMenu->setItemChecked( sortMenu->idAt( 2 ), FALSE );
    sortMenu->setItemChecked( sortMenu->idAt( 3 ), FALSE );
}

void FileBrowser::sortDate()
{
    fileView->setSorting( 1 );
    fileView->sort();
    sortMenu->setItemChecked( sortMenu->idAt( 0 ), FALSE );
    sortMenu->setItemChecked( sortMenu->idAt( 1 ), TRUE );
    sortMenu->setItemChecked( sortMenu->idAt( 2 ), FALSE );
    sortMenu->setItemChecked( sortMenu->idAt( 3 ), FALSE );
}

void FileBrowser::sortSize()
{
    fileView->setSorting( 2 );
    fileView->sort();
    sortMenu->setItemChecked( sortMenu->idAt( 0 ), FALSE );
    sortMenu->setItemChecked( sortMenu->idAt( 1 ), FALSE );
    sortMenu->setItemChecked( sortMenu->idAt( 2 ), TRUE );
    sortMenu->setItemChecked( sortMenu->idAt( 3 ), FALSE );
}

void FileBrowser::sortType()
{
    fileView->setSorting( 3 );
    fileView->sort();
    sortMenu->setItemChecked( sortMenu->idAt( 0 ), FALSE );
    sortMenu->setItemChecked( sortMenu->idAt( 1 ), FALSE );
    sortMenu->setItemChecked( sortMenu->idAt( 2 ), FALSE );
    sortMenu->setItemChecked( sortMenu->idAt( 3 ), TRUE );
}
