/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "organizer.h"
#include <qdatetime.h>

static int last_uid = 0;
static int get_uid()
{
    return last_uid++;
}



OrganizerBase::OrganizerBase()
{
    id = get_uid();
}

int OrganizerBase::uid() const
{
    return id;
}





Note::Note()
{
}

void Note::setText( const QString &text )
{
    txt = text;
}

void Note::setPixmap( const QPixmap &pm )
{
    pix = pm;
}

QString Note::text() const
{
    return txt;
}

QPixmap Note::pixmap() const
{
    return pix;
}





Todo::Todo()
    : OrganizerBase(), hDate( FALSE ), dt( QDate::currentDate() ),
      complete( FALSE ), prior( 6 ), descript(), categ(), note()
{
}

void Todo::setDate( bool b )
{
    hDate = b;
}

void Todo::setDate( const QDate &dt )
{
    this->dt = dt;
}

bool Todo::hasDate() const
{
    return hDate;
}

QDate Todo::date() const
{
    return dt;
}

void Todo::setCompleted( bool b )
{
    complete = b;
}

bool Todo::isCompleted() const
{
    return complete;
}

void Todo::setPriority( int p )
{
    prior = p;
}

int Todo::priority() const
{
    return prior;
}

void Todo::setDescription( const QString &d )
{
    descript = d;
}

QString Todo::description() const
{
    return descript;
}

void Todo::setCategory( const QString &c )
{
    categ = c;
}

QString Todo::category() const
{
    return categ;
}

void Todo::setNotes( const Note &n )
{
    note = n;
}

Note Todo::notes() const
{
    return note;
}

static void write( QTextStream &ts, const QDate& dt )
{
    ts << "<date><year>" << dt.year() << "</year><month>" << dt.month() << "</month><day>" << dt.day() << "</day></date>";
}

static void write( QTextStream &ts, const QTime& dt )
{
    ts << "<time><hour>" << dt.hour() << "</hour><minute>" << dt.minute() << "</minute><second>" << dt.second() << "</second></time>";
}

static void write( QTextStream &ts, const QDateTime& dt )
{
    write(ts,dt.date());
    write(ts,dt.time());
}

static void read( Node* n, QDate& date )
{
    int y=0, m=0, d=0;
    while ( n ) {
	if ( n->tagName() == "year" )
	    y = n->data().toInt();
	else if ( n->tagName() == "month" )
	    m = n->data().toInt();
	else if ( n->tagName() == "day" )
	    d = n->data().toInt();
	n = n->nextNode();
    }
    date.setYMD( y, m, d );
}

static void read( Node* n, QTime& time )
{
    int h=0, m=0, s=0;
    while ( n ) {
	if ( n->tagName() == "hour" )
	    h = n->data().toInt();
	else if ( n->tagName() == "minute" )
	    m = n->data().toInt();
	else if ( n->tagName() == "second" )
	    s = n->data().toInt();
	n = n->nextNode();
    }
    time.setHMS( h, m, s );
}

static void read( Node* n, QDateTime& dt )
{
    while ( n ) {
	if ( n->tagName() == "date" ) {
	    QDate d;
	    read(n->firstChild(),d);
	    dt.setDate(d);
	} else if ( n->tagName() == "time" ) {
	    QTime t;
	    read(n->firstChild(),t);
	    dt.setTime(t);
	}
	n = n->nextNode();
    }
}

void Todo::save( QTextStream &ts, const QString &indent )
{
    ts << indent << "<hasdate>" << (int)hDate << "</hasdate>" << endl;
    ts << indent; write(ts,dt); ts << endl;
    ts << indent << "<completed>" << (int)complete << "</completed>" << endl;
    ts << indent << "<priority>" << prior << "</priority>" << endl;
    ts << indent << "<description>" << descript << "</description>" << endl;
    ts << indent << "<category>" << categ << "</category>" << endl;
    ts << indent << "<note>" << note.text() << "</note>" << endl;
}

void Todo::load( Node *n )
{
    while ( n ) {
	if ( n->tagName() == "hasdate" ) {
	    hDate = (bool)n->data().toInt();
	} else if ( n->tagName() == "date" ) {
	    read(n->firstChild(),dt);
	} else if ( n->tagName() == "completed" ) {
	    complete = (bool)n->data().toInt();
	} else if ( n->tagName() == "priority" ) {
	    prior = n->data().toInt();
	} else if ( n->tagName() == "description" ) {
	    descript = n->data();
	} else if ( n->tagName() == "category" ) {
	    categ = n->data();
	} else if ( n->tagName() == "note" ) {
	    note.setText( n->data() );
	}
	n = n->nextNode();
    }
}


Event::Event()
{
    startDate = endDate = QDateTime::currentDateTime();
    typ = Normal;
    hAlarm = FALSE;
    hRepeat = FALSE;
    aMinutes = 0;
    // pattern = something
}

void Event::setDescription( const QString &s )
{
    descript = s;
}

void Event::setLocation( const QString &s )
{
    locat = s;
}

void Event::setCategory( const QString &s )
{
    categ = s;
}

void Event::setType( Type t )
{
    typ = t;
}

void Event::setStart( const QDateTime &d )
{
    startDate = d;
}

void Event::setEnd( const QDateTime &d )
{
    endDate = d;
}

void Event::setAlarm( bool b, int minutes )
{
    hAlarm = b;
    aMinutes = minutes;
}

void Event::setRepeat( bool b, const RepeatPattern &p )
{
    hRepeat = b;
    pattern = p;
}

void Event::setNotes( const QString &n )
{
    note = n;
}

QString Event::description() const
{
    return descript;
}

QString Event::location() const
{
    return locat;
}

QString Event::category() const
{
    return categ;
}

Event::Type Event::type() const
{
    return typ;
}

QDateTime Event::start( bool actual ) const
{
    QDateTime dt = startDate;
    if ( actual && typ == AllDay ) {
	QTime t = dt.time();
	t.setHMS( 0, 0, 0 );
	dt.setTime( t );
    }
    return dt;
}

QDateTime Event::end( bool actual ) const
{
    QDateTime dt = endDate;
    if ( actual && typ == AllDay ) {
	QTime t = dt.time();
	t.setHMS( 23, 59, 59 );
	dt.setTime( t );
    }
    return dt;
}

bool Event::hasAlarm() const
{
    return hAlarm;
}

int Event::alarmTime() const
{
    return aMinutes;
}

bool Event::doRepeat() const
{
    return hRepeat;
}

Event::RepeatPattern Event::repeatPattern() const
{
    return pattern;
}

QString Event::notes() const
{
    return note;
}

bool Event::operator==( const Event &e ) const
{
    return ( e.descript == descript &&
	     e.locat == locat &&
	     e.categ == categ &&
	     e.typ == typ &&
	     e.startDate == startDate &&
	     e.endDate == endDate &&
	     e.hAlarm == hAlarm &&
	     e.hRepeat == hRepeat &&
	     e.aMinutes == aMinutes &&
	     e.pattern == pattern &&
	     e.note == note );
}

void Event::save( QTextStream &ts, const QString &indent )
{
    ts << indent << "<description>" << descript << "</description>" << endl;
    ts << indent << "<location>" << locat << "</location>" << endl;
    ts << indent << "<category>" << categ << "</category>" << endl;
    if ( typ != Normal ) {
	ts << indent << "<type>";
	switch ( typ ) {
	    case Normal:
		break;
	    case AllDay:
		ts << "AllDay";
	}
	ts << "</type>" << endl;
    }
    if ( hAlarm ) {
	ts << indent << "<alarm>" << aMinutes << "</alarm>" << endl;
    }
    if ( hRepeat ) {
	ts << indent << "<repeat>";
	    // ## something
	    ts << "</repeat>" << endl;
    }
    ts << indent << "<start>";
	write(ts,startDate);
	ts << "</start>" << endl;
    if ( startDate != endDate ) {
	ts << indent << "<end>";
	    if ( startDate.date() != endDate.date() )
		write(ts,endDate.date());
	    write(ts,endDate.time());
	    ts << "</end>" << endl;
    }
    if ( !note.isEmpty() ) {
	ts << indent << "<note>" << note << "</note>" << endl;
    }
}

void Event::load( Node *n )
{
    while ( n ) {
	if ( n->tagName() == "description" ) {
	    descript = n->data();
	} else if ( n->tagName() == "start" ) {
	    read(n->firstChild(), startDate);
	} else if ( n->tagName() == "end" ) {
	    read(n->firstChild(), endDate);
	} else if ( n->tagName() == "location" ) {
	    locat = n->data();
	} else if ( n->tagName() == "category" ) {
	    categ = n->data();
	} else if ( n->tagName() == "type" ) {
	    if ( n->data() == "AllDay" )
		typ = AllDay;
	} else if ( n->tagName() == "alarm" ) {
	    categ = n->data().toInt();
	} else if ( n->tagName() == "repeat" ) {
	    // ## something
	} else if ( n->tagName() == "note" ) {
	    note = n->data();
	}
	n = n->nextNode();
    }
}


