/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#ifndef __CARD_GAME_H__
#define __CARD_GAME_H__


#include <qpoint.h>
#include <qlist.h>


enum eSuit {
    jokerSuit = 0, clubs, spades, diamonds, hearts
};


enum eValue {
    jokerVal = 0, ace, two, three, four, five,
    six, seven, eight, nine, ten, jack, queen, king
};


enum ePileStackingType {
    pileCascades = 0, pileStacks, pileCascadesOrStacks
};


enum ePileFaceingType {
    pileFaceUp = 0, pileFaceDown, pileFaceUpOrDown
};


class CardPile;


class Card
{
public:
    Card( eValue v, eSuit s, bool f ) :
    	val(v), suit(s), faceUp(f), showing(FALSE), ix(0), iy(0), iz(0), cardPile(NULL) { }
    virtual ~Card() { }
    eValue getValue() { return val; }
    eSuit getSuit() { return suit; }
    CardPile *getCardPile() { return cardPile; }
    bool isFacing() { return faceUp; }
    bool isShowing() { return showing; }
    bool isRed() { return ((suit == diamonds) || (suit == hearts)); }
    int getX(void) { return ix; }
    int getY(void) { return iy; }
    int getZ(void) { return iz; }
    void setCardPile(CardPile *p) { cardPile = p; }
    void setFace(bool f) { faceUp = f; /* flip(); */ }
    void flip(void) { flipTo(getX(), getY()); }
    virtual void setPos(int x, int y, int z) { ix = x; iy = y; iz = z; }
    virtual void move(int x, int y) { ix = x; iy = y; }
    virtual void move(QPoint p) { ix = p.x(); iy = p.y(); }
    virtual void flipTo(int x, int y, int steps = 8) { ix = x; iy = y; faceUp = !faceUp; redraw(); Q_UNUSED(steps); }
    virtual void showCard(void) { showing = TRUE; }
    virtual void hideCard(void) { showing = FALSE; }
protected:
    virtual void redraw(void) { }
private:
    eValue val;
    eSuit suit;
    bool faceUp;
    bool showing;
    int ix, iy, iz;
    CardPile *cardPile;
};


class CardDeck
{
public:
    CardDeck(int jokers = 0) : numberOfJokers(jokers) {
	cards = new (Card *)[getNumberOfCards()];
    }
    virtual ~CardDeck() {
	for (int i = 0; i < getNumberOfCards(); i++)
	    delete cards[i];
	delete cards;
    }
    void createDeck(void);
    void shuffle(void);
    int getNumberOfCards(void) {  return 52 + getNumberOfJokers(); }
    int getNumberOfJokers(void) {  return numberOfJokers; }
    
    virtual Card *newCard( eValue v, eSuit s, bool f ) { return new Card(v, s, f); }
    virtual void deal(void) { }

    Card **cards;
private:
    int numberOfJokers;
};


class CardPile : public QList<Card>
{
public:
    CardPile(int x, int y) : pileX(x), pileY(y), dealing(FALSE) {
	pileWidth = 0;
	pileHeight = 0;
	pileNextX = pileX;
	pileNextY = pileY;
	pileCenterX = x + pileWidth / 2;
	pileCenterY = y + pileHeight / 2;
	pileRadius = (pileWidth > pileHeight) ? pileWidth : pileHeight;
    }
    
    int getX(void) { return pileX; }
    int getY(void) { return pileY; }
    int getNextX(void) { return pileNextX; }
    int getNextY(void) { return pileNextY; }
    int getWidth(void) { return pileWidth; }
    int getHeight(void) { return pileHeight; }

    void setX(int x) { pileX = x; }
    void setY(int y) { pileY = y; }
    void setNextX(int x) { pileNextX = x; }
    void setNextY(int y) { pileNextY = y; }
    void setWidth(int width) { pileWidth = width; }
    void setHeight(int height) { pileHeight = height; }
    
    void beginDealing(void) { dealing = TRUE; }
    void endDealing(void) { dealing = FALSE; }
    bool isDealing(void) { return dealing; }
    
    int distanceFromPile(int x, int y) { return (pileCenterX-x)*(pileCenterX-x)+(pileCenterY-y)*(pileCenterY-y); }
    int distanceFromNextPos(int x, int y) { return (pileNextX-x)*(pileNextX-x)+(pileNextY-y)*(pileNextY-y); }
    
    Card *cardOnTop(void) {
	return getLast(); 
    }
    Card *cardOnBottom(void) {
	return getFirst(); 
    }
    Card *cardInfront(Card *c) {
	CardPile *p = c->getCardPile();
	if (p) {
	    p->at(p->find(c));
	    return p->next();
	} else {
	    return NULL;
	}
    }

    bool kingOnTop() {
	Card *top = cardOnTop();
	return top && top->getValue() == king;
    }
    
    bool addCardToTop(Card *c) {
	if (dealing || isAllowedOnTop(c)) {
    	    append((const Card *)c);
	    cardAddedToTop(c);
	    return TRUE;
	}
	return FALSE;
    }
    bool addCardToBottom(Card *c) {
	if (dealing || isAllowedOnBottom(c)) {
    	    prepend((const Card *)c);
	    cardAddedToBottom(c);
	    return TRUE;
	}
	return FALSE;
    }
    bool removeCard(Card *c) {
	if (dealing || isAllowedToBeMoved(c)) {
	    take(find(c));
	    cardRemoved(c);
	    return TRUE;
	}
	return FALSE;
    }
    
    virtual void cardAddedToTop(Card *card) {
	Q_UNUSED(card);    
    }
    virtual void cardAddedToBottom(Card *card) {
	Q_UNUSED(card);    
    }
    virtual void cardRemoved(Card *card) {
	Q_UNUSED(card);    
    }
    virtual bool isAllowedOnTop(Card *card) {
	Q_UNUSED(card);    
	return FALSE;
    }
    virtual bool isAllowedOnBottom(Card *card) {
	Q_UNUSED(card);    
	return FALSE;
    }
    virtual bool isAllowedToBeMoved(Card *card) {
	Q_UNUSED(card);    
	return FALSE;
    }
    virtual QPoint getCardPos(Card *c) {
	Q_UNUSED(c);
	return QPoint(pileX, pileY); 
    }
    virtual QPoint getHypertheticalNextCardPos(void) {
	return QPoint(pileX, pileY); 
    }
    
protected:    
    int pileX, pileY;
    int pileNextX, pileNextY;
    int pileWidth, pileHeight;
    int pileCenterX, pileCenterY;
    int pileRadius;
private:
    bool dealing;
};


class CardGameLayout : public QList<CardPile>
{
public:
    CardGameLayout() { }
    virtual ~CardGameLayout() {
	for (CardPile *p = first(); p != NULL; p = next())
	    delete p;
    }
    
    void addCardPile(CardPile *pile) {
	append((const CardPile *)pile);
    }
    CardPile *closestPile(int x, int y, int maxDistance) {
	int closestDistance = maxDistance * maxDistance;
	CardPile *closestPile = NULL;
	
	for (CardPile *p = first(); p != NULL; p = next()) {
    	    int d = p->distanceFromNextPos(x, y);
	    if (d < closestDistance) {
		closestDistance = d;
		closestPile = p;		
	    }
	}
	
	return closestPile;
    }
    void beginDealing(void) {
	for (CardPile *p = first(); p != NULL; p = next())
	    p->beginDealing();
    }
    void endDealing(void) {
	for (CardPile *p = first(); p != NULL; p = next())
	    p->endDealing();
    }
private:
};


class CardGame : public CardGameLayout, public CardDeck
{
public:
    CardGame(int numOfJokers = 0) : CardGameLayout(), CardDeck(numOfJokers), deckCreated(FALSE) { }
    virtual void newGame(void) {
	// Create Cards
	if (!deckCreated) {
    	    createDeck();
	    deckCreated = TRUE;
	}

	// Shuffle Cards
	shuffle();
     
	// Deal Cards
	deal();
    }
    virtual void mousePress(QPoint p) { Q_UNUSED(p); }
    virtual void mouseRelease(QPoint p) { Q_UNUSED(p); }
    virtual void mouseMove(QPoint p) { Q_UNUSED(p); }
private:
    bool deckCreated;
};


#endif

