/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "todotable.h"

#include <qpainter.h>
#include <qcombobox.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qtextcodec.h>
#include <qxml.h>

#include <xmlreader.h>

CheckItem::CheckItem( QTable *t )
    : QTableItem( t, Never, "" ), checked( FALSE )
{
}

void CheckItem::setChecked( bool b )
{
    checked = b;
    table()->updateCell( row(), col() );
}

void CheckItem::toggle()
{
    checked = !checked;
    table()->updateCell( row(), col() );
}

bool CheckItem::isChecked() const
{
    return checked;
}

static const int BoxSize = 10;

void CheckItem::paint( QPainter *p, const QColorGroup &cg, const QRect &cr, bool )
{
    p->fillRect( 0, 0, cr.width(), cr.height(), cg.brush( QColorGroup::Base ) );

    int marg = ( cr.width() - BoxSize ) / 2;
    int x = 0;
    int y = ( cr.height() - BoxSize ) / 2;
    p->setPen( QPen( cg.text() ) );
    p->drawRect( x + marg, y, BoxSize, BoxSize );
    p->drawRect( x + marg+1, y+1, BoxSize-2, BoxSize-2 );
    p->setPen( darkGreen );
    x += 1;
    y += 1;
    if ( checked ) {
	QPointArray a( 7*2 );
	int i, xx, yy;
	xx = x+1+marg;
	yy = y+2;
	for ( i=0; i<3; i++ ) {
	    a.setPoint( 2*i,   xx, yy );
	    a.setPoint( 2*i+1, xx, yy+2 );
	    xx++; yy++;
	}
	yy -= 2;
	for ( i=3; i<7; i++ ) {
	    a.setPoint( 2*i,   xx, yy );
	    a.setPoint( 2*i+1, xx, yy+2 );
	    xx++; yy--;
	}
	p->drawLineSegments( a );
    }
}


ComboItem::ComboItem( QTable *t, EditType et )
    : QTableItem( t, et, "3" ), cb( 0 )
{
    setReplaceable( FALSE );
}

QWidget *ComboItem::createEditor() const
{
    QString txt = text();
    ( (ComboItem*)this )->cb = new QComboBox( table()->viewport() );
    cb->insertItem( "1" );
    cb->insertItem( "2" );
    cb->insertItem( "3" );
    cb->insertItem( "4" );
    cb->insertItem( "5" );
    cb->setCurrentItem( txt.toInt() - 1 );
    return cb;
}

void ComboItem::setContentFromEditor( QWidget *w )
{
    if ( w->inherits( "QComboBox" ) )
	setText( ( (QComboBox*)w )->currentText() );
    else
	QTableItem::setContentFromEditor( w );
}	

void ComboItem::setText( const QString &s )
{
    if ( cb )
	cb->setCurrentItem( s.toInt() - 1 );
    QTableItem::setText( s );
}

QString ComboItem::text() const
{
    if ( cb )
	return cb->currentText();
    return QTableItem::text();
}



TodoTable::TodoTable( QWidget *parent, const char *name )
    : QTable( 0, 3, parent, name )
{
    setSorting( TRUE );
    setSelectionMode( NoSelection );
    setColumnStretchable( 2, TRUE );
    setColumnWidth( 0, 20 );
    setColumnWidth( 1, 35 );
    setLeftMargin( 0 );
    verticalHeader()->hide();
    horizontalHeader()->setLabel( 0, tr( "C." ) );
    horizontalHeader()->setLabel( 1, tr( "Prior." ) );
    horizontalHeader()->setLabel( 2, tr( "Description" ) );
    connect( this, SIGNAL( pressed( int, int, int, const QPoint & ) ),
	     this, SLOT( slotClicked( int, int, int, const QPoint & ) ) );
}

void TodoTable::addEntry( Todo *todo )
{
    int row = numRows();
    setNumRows( row + 1 );
    CheckItem *ci = new CheckItem( this );
    setItem( row, 0, ci );
    ci->setChecked( todo->isCompleted() );
    ComboItem *co = new ComboItem( this, QTableItem::WhenCurrent );
    setItem( row, 1, co );
    co->setText( QString::number( todo->priority() ) );

    QTableItem *ti = new QTableItem( this, QTableItem::WhenCurrent, todo->description() );
    ti->setReplaceable( FALSE );
    setItem( row, 2, ti );

    todoList.insert( ci, todo );
}

void TodoTable::slotClicked( int row, int col, int, const QPoint &pos )
{
    if ( col != 0 )
	return;
    CheckItem *i = (CheckItem*)item( row, col );
    int x = pos.x() - columnPos( col );
    int y = pos.y() - rowPos( row );
    int w = columnWidth( col );
    int h = rowHeight( row );
    if ( i && x >= ( w - BoxSize ) / 2 && x <= ( w - BoxSize ) / 2 + BoxSize &&
	 y >= ( h - BoxSize ) / 2 && y <= ( h - BoxSize ) / 2 + BoxSize ) {
	i->toggle();
    }
}

Todo *TodoTable::currentEntry() const
{
    QTableItem *i = item( currentRow(), 0 );
    if ( !i )
	return 0;
    Todo *todo = *todoList.find( (CheckItem*)i );
    todo->setCompleted( ( (CheckItem*)item( currentRow(), 0 ) )->isChecked() );
    todo->setPriority( ( (ComboItem*)item( currentRow(), 1 ) )->text().toInt() );
    todo->setDescription( text( currentRow(), 2 ) );
    return todo;
}

void TodoTable::replaceEntry( Todo *todo )
{
    int row = currentRow();
    todoList.remove( (CheckItem*)item( row, 0 ) );
    ( (CheckItem*)item( row, 0 ) )->setChecked( todo->isCompleted() );
    ( (ComboItem*)item( row, 1 ) )->setText( QString::number( todo->priority() ) );
    item( row, 2 )->setText( todo->description() );
    todoList.insert( (CheckItem*)item( row, 0 ), todo );
}

void TodoTable::save( const QString &fn )
{
    QFile f( fn );
    if ( !f.open( IO_WriteOnly ) )
	return;
    QTextStream ts( &f );
    ts.setCodec( QTextCodec::codecForName( "UTF-8" ) );
    ts << "<!DOCTYPE TODO><TODO>" << endl;
    for ( QMap<CheckItem*, Todo *>::Iterator it = todoList.begin(); it != todoList.end(); ++it ) {
	if ( !item( it.key()->row(), 0 ) )
	    continue;
	Todo *todo = *it;
	ts << "<task>" << endl;
	todo->save( ts, "    " );
	ts << "</task>" << endl;
    }
    ts << "</TODO>" << endl;
}


void TodoTable::load( const QString &fn )
{
    QFile f( fn );
    if ( !f.open( IO_ReadOnly ) )
	return;

    QTextStream ts( &f );
    QXmlInputSource inputsource( ts );
    QXmlSimpleReader reader;
    XmlHandler handler;

    reader.setFeature( "http://trolltech.com/xml/features/report-whitespace-only-CharData", FALSE );
    reader.setContentHandler( &handler );
    reader.parse( inputsource );

    Node *n = handler.firstNode();

    while ( n ) {
	if ( n->tagName() == "task" ) {
	    Todo *todo = new Todo;
	    todo->load( n->firstChild() );
	    addEntry( todo );
	    n = n->nextNode();
	    continue;
	}
	if ( n->lastChild() ) {
	    n = n->firstChild();
	} else {
	    if ( !n->nextNode() && n->parentNode() )
		n = n->parentNode();
	    n = n->nextNode();
	}
    }
}

void TodoTable::clear()
{
    todoList.clear();
    for ( int r = 0; r < numRows(); ++r ) {
	for ( int c = 0; c < numCols(); ++c ) {
	    clearCell( r, c );
	}
    }
    setNumRows( 0 );
}
